import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from sklearn.metrics import *
from sklearn.cluster import KMeans
from sklearn.cluster import DBSCAN
from sklearn.cluster import AgglomerativeClustering
from sklearn.neighbors import kneighbors_graph
from sklearn import preprocessing

from scipy.stats import mode
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import linkage, dendrogram

__author__ = 'Riccardo Guidotti'


def categorical_to_numeric(categories):
    category_to_number = dict()
    number_to_category = dict()

    index = 0
    for c in categories:
        category_to_number[c] = index
        number_to_category[index] = c

    return category_to_number, number_to_category


def main():

    input_filename = '/Users/riccardo/Documents/Didattica/CorsoDM_2015_2016/python-workspace/adult.csv'
    df = pd.read_csv(input_filename, skipinitialspace=True)

    df = df[0:1000]

    # remove useless features
    df = df.drop('class', 1)
    df = df.drop('education', 1)

    # fill missing values
    for k in df.columns:
        mode_value = mode(df[k])[0][0]
        df[k] = df[k].fillna(mode_value)

    # from categorical to numerical
    features_to_categorize = ['workclass', 'marital-status', 'occupation', 'relationship', 'race', 'sex',
                              'native-country']
    maps_cn = dict()
    for f in features_to_categorize:
        ctn, ntc = categorical_to_numeric(df[f])
        maps_cn[f] = (ctn, ntc)
        df[f] = df[f].map(ctn).astype(int)

    # normalize values betwen 0 and 1
    min_max_scaler = preprocessing.MinMaxScaler()
    for k in df.columns:
        df[k] = min_max_scaler.fit_transform(df[k].values.astype(float))

    n_clusters = 3

    # centroid-based clustering
    # init 'random', 'k-means++'

    print '\nkmeans'
    kmeans = KMeans(init='k-means++', n_clusters=n_clusters, n_init=10, max_iter=100)
    kmeans.fit(df.values)

    hist, bins = np.histogram(kmeans.labels_, bins=range(0, len(set(kmeans.labels_)) + 1))

    print 'centers', kmeans.cluster_centers_
    print 'labels', dict(zip(bins, hist))
    print 'sse', kmeans.inertia_
    print 'silhouette', silhouette_score(df.values, kmeans.labels_)

    # hierarchical clustering
    connectivity = kneighbors_graph(df.values, n_neighbors=100, include_self=False)
    connectivity = 0.5 * (connectivity + connectivity.T)

    print '\nward'
    ward = AgglomerativeClustering(n_clusters=n_clusters, linkage='ward', affinity='euclidean',
                                   connectivity=connectivity)
    ward.fit(df.values)

    hist, bins = np.histogram(ward.labels_, bins=range(0, len(set(ward.labels_)) + 1))

    print 'labels', dict(zip(bins, hist))
    print 'silhouette', silhouette_score(df.values, ward.labels_)

    print '\naverage linkage'
    average_linkage = AgglomerativeClustering(n_clusters=n_clusters, linkage='average', affinity='manhattan',
                                              connectivity=connectivity)
    average_linkage.fit(df.values)

    hist, bins = np.histogram(average_linkage.labels_, bins=range(0, len(set(average_linkage.labels_)) + 1))

    print 'labels', dict(zip(bins, hist))
    print 'silhouette', silhouette_score(df.values, average_linkage.labels_)

    print '\ncomplete linkage'
    complete_linkage = AgglomerativeClustering(n_clusters=n_clusters, linkage='complete', affinity='l1',
                                              connectivity=connectivity)
    complete_linkage.fit(df.values)
    hist, bins = np.histogram(complete_linkage.labels_, bins=range(0, len(set(complete_linkage.labels_)) + 1))

    print 'labels', dict(zip(bins, hist))
    print 'silhouette', silhouette_score(df.values, complete_linkage.labels_)

    # density based clustering
    print '\ndbscan'

    dbscan = DBSCAN(eps=0.1, min_samples=5, metric='euclidean')
    dbscan.fit(df.values)

    hist, bins = np.histogram(dbscan.labels_, bins=range(-1, len(set(dbscan.labels_)) + 1))

    print 'labels', dict(zip(bins, hist))
    print 'silhouette', silhouette_score(df.values, dbscan.labels_)

    # single linkage and dendogram using scipy
    data_dist = pdist(df.values, metric='euclidean')
    data_link = linkage(data_dist, method='single', metric='euclidean')
    dendrogram(data_link)
    plt.show()


if __name__ == "__main__":
    main()

